/*******************************************************************************
Copyright © 2023 Sebastian Roith

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*******************************************************************************/

(function(window) {

const level = {
    // Angezeigter Arbeitsauftrag (HTML erlaubt)
    task: 'Der Roboter muss dringend eine Mauer restaurieren und die fehlenden Plättchen ersetzen. Entwirf einen Algorithmus, der alle Lücken auffüllt. Nutze dabei geeignete selbstdefinierte Anweisungen.',

    // Welt und Roboterstartposition als Maschinencode
    world: '03',

    // Quelltext, der anfangs im Editor steht
    code: '# Selbstdefinierte Anweisungen\n\n# Alle Lücken der konkreten Mauer auffüllen\n',

    // Ist das Bearbeiten des Quelltextes gesperrt? (Optional, Standardwert false)
    readonly: false,

    // Soll der Quellcode nicht geändert werden, wenn zuvor das direkt vorherige Level aktiv war?
    keepCode: false,

    // Muss - ggf. zusätzlich zur korrekten Fragenbeantwortung - checkCode oder checkTermination unlock=true liefern, bevor Level abgeschlossen werden kann? (Optional, Standardwert false)
    locked: true,

    // Optional: Methode, die beim Start des Levels ausgeführt werden soll
    start() { },

    // Optional: Überprüft, ob ein syntaktisch korrekter Quellcode ausgeführt werden darf; Rückgabewert true für ja, sonst { accept: "Ausführung erlauben?", line: "Zeilennummer eines Fehlers (optional)", msg: "Fehlermeldung (optional)", fine: "Als einen Fehlversuch werten? (optional)", unlock: "Level als abgeschlossen behandeln? (optional)" }
    checkCode: [
        CodeCheckers.countCommand(0, 'minus', 0, 0, '„minus“ ist zum Auffüllen ungeeignet…'),
        CodeCheckers.accept()
    ],

    // Optional: Reagiert auf ein Ende der Algorithmusausführung; Rückgabewert wie checkCode
    checkTermination: [
        CodeCheckers.acceptUserAbort(),
        CodeCheckers.allowOnlyDone(1),
        false,
        CodeCheckers.countUserDefined(1, 2),
        CodeCheckers.countUserDefinedUnused(1, 0),
        CodeCheckers.countUserDefinedCalls(1, 3),
        CodeCheckers.accept(true)
    ],

    // Wird falsches Beantworten der Frage als Fehlversuch gewertet? (Optional, Standardwert false)
    questionWrongFine: 0,

    // Falls erst eine Frage richtig beantwortet werden muss: Welche?
    question: null
};

function getUniqueValues(fktRandomInt, count, min, max) {
    const values = [];
    while (values.length < count) {
        const randomValue = fktRandomInt(min, max - values.length);
        values.push(values.slice().sort((a, b) => a - b).reduce((newValue, oldValue) => ((newValue >= oldValue) ? (newValue + 1) : newValue), randomValue));
    }
    return values;
}

function genLevel(fktRandomInt) {
    const randomLevel = Object.create(level);
    const boolHasWidth = fktRandomInt(0, 1);
    const length = fktRandomInt(9, 13);
    randomLevel.world += boolHasWidth ? (length.toString(16) + '1') : ('1' + length.toString(16));
    const depthOfHoles = getUniqueValues(fktRandomInt, 3, 1, 3).sort().map((v, index) => v + index);
    const maxDepthOfHoles = depthOfHoles[depthOfHoles.length - 1];
    const wallHeight = fktRandomInt(1, 15 - maxDepthOfHoles) + maxDepthOfHoles;
    const resultChecker = CodeCheckers.worldCheck(1, (new Array(length).fill(wallHeight)));
    randomLevel.checkTermination = randomLevel.checkTermination.map(checker => (checker === false) ? resultChecker : checker);

    const countHoles = fktRandomInt(5, 8);
    const holeIndizes = getUniqueValues(fktRandomInt, countHoles, 0, length - 1);
    const countHoleTypes = depthOfHoles.map(() => 0);
    const plates = new Array(length);
    plates.fill(wallHeight);
    holeIndizes.map((value, index) => {
        const remaining = countHoles - 1 - index;
        if ((remaining < depthOfHoles.length) && !(countHoleTypes[remaining])) {
            plates[value] -= depthOfHoles[remaining];
            return;
        }
        const depthIndex = fktRandomInt(0, depthOfHoles.length - 1);
        ++countHoleTypes[depthIndex];
        plates[value] -= depthOfHoles[depthIndex];
    });
    randomLevel.world += plates.map(v => ('0' + v.toString(16))).join('');
    const robotPos = (fktRandomInt(0, 1) * (length - 1)).toString(16);
    randomLevel.world += (boolHasWidth ? (robotPos + '0') : ('0' + robotPos));
    randomLevel.world += fktRandomInt(0, 3);
    return [ randomLevel ];
}

window.roboterCodingTask = {
    levels: genLevel
};
})(window);
